@@ -30,10 +30,10 @@ class DataLoader:
     def __init__(
         self,
         mode,
+        data_path,
         imgA_label=None,
         imgB_label=None,
-        input_size=None,
-        data_dir="./Data",
+        img_size=None,
         output_dir=None,
         is_B_categorical=False,
         num_classes=None,
@@ -41,6 +41,8 @@ class DataLoader:
         norm_boundsB=None,
         extract_only=None,
         use_3D=False,
+        patch_size=None,
+        patch_overlap=0.0,
     ):
         """[summary]
 
@@ -52,8 +54,8 @@ class DataLoader:
             imgB_label (str): Identifier for class B. It's the name of the
                 folder inside :py:attr:`data_dir` that contains images
                 labeled as class B.
-            input_size (int): Dimension of a single image, defined as
-                input_size x input_size. Currently, it supports only squared
+            img_size (int): Dimension of a single image, defined as
+                img_size x img_size. Currently, it supports only squared
                 images.
             data_dir (str, optional): Path to directory that contains the
                 Dataset. This folder **must** contain two subfolders named like
@@ -92,47 +94,63 @@ class DataLoader:
         self.mode = mode
 
         if mode == "gen":
-            if not os.path.exists(data_dir):
-                raise FileNotFoundError(f"{data_dir} does not exist")
+            dir_mode = False
+            file_mode = False
 
-            self.data_dir = data_dir
+            if os.path.isdir(data_path):
+                if not os.path.exists(data_path):
+                    raise FileNotFoundError(f"{data_path} does not exist")
+                else:
+                    dir_mode = True
+            else:
+                if not os.path.exists(data_path):
+                    raise FileNotFoundError(f"{data_path} does not exist")
+                else:
+                    file_mode = True
+
+            self.data_path = data_path
 
             if output_dir is None:
-                self.output_dir = os.path.join(os.path.dirname(data_dir),
-                                               f"{os.path.basename(data_dir)}_TF")  # noqa
+                if dir_mode:
+                    self.output_dir = os.path.join(os.path.dirname(self.data_path),
+                                                   f"{os.path.basename(self.data_path)}_TF")
+                elif file_mode:
+                    base_dirname = os.path.dirname(self.data_path)
+                    output_dirname = os.path.basename(
+                        self.data_path).replace(".json", "_TF")
+                    self.output_dir = os.path.join(
+                        base_dirname, output_dirname)
             else:
                 self.output_dir = output_dir
             os.makedirs(self.output_dir, exist_ok=True)
 
-            if input_size is None:
-                raise ValueError("input_size is None")
-
-            self.input_size = input_size
-
-            if (not isinstance(use_3D, bool)):
-                raise ValueError("use_3D is not a Boolean value")
-            self.use_3D = use_3D
-
             if imgA_label is None or imgB_label is None:
                 raise ValueError("imgA_label or imgB_label is None.")
 
             self.imgA_label = imgA_label
             self.imgB_label = imgB_label
 
-            if not os.path.exists(os.path.join(data_dir,
-                                               imgA_label)):
-                raise FileNotFoundError(f"{imgA_label} does not exist")
+            if dir_mode:
+                if not os.path.exists(os.path.join(data_path,
+                                                   imgA_label)):
+                    raise FileNotFoundError(f"{imgA_label} does not exist")
 
-            if not os.path.exists(os.path.join(data_dir,
-                                               imgB_label)):
-                raise FileNotFoundError(f"{imgB_label} does not exist")
+                if not os.path.exists(os.path.join(data_path,
+                                                   imgB_label)):
+                    raise FileNotFoundError(f"{imgB_label} does not exist")
 
-            self.imgA_paths, self.imgB_paths = self.get_imgs_paths()
+                self.imgA_paths, self.imgB_paths = self.get_imgs_paths()
+            elif file_mode:
+                self.imgA_paths, self.imgB_paths = self.read_imgs_paths()
 
             if extract_only is not None:
                 self.imgA_paths = self.imgA_paths[:extract_only]
                 self.imgB_paths = self.imgB_paths[:extract_only]
 
+            if not isinstance(use_3D, bool):
+                raise ValueError("use_3D is not a Boolean value")
+            self.use_3D = use_3D
+
             self.is_3D = self.is_3D_data(self.imgA_paths[0])
             if self.is_3D:
                 self.is_A_RGB = self.is_RGB_data(self.imgA_paths[0])
@@ -147,6 +165,32 @@ class DataLoader:
                 raise ValueError(
                     "Image files are not 3D but use_3D was set to True")
 
+            if img_size is None:
+                raise ValueError("img_size is None")
+            elif isinstance(img_size, int):
+                img_size = [img_size]
+                if self.is_3D and self.use_3D:
+                    img_size = img_size * 3
+                else:
+                    img_size = img_size * 2
+            elif isinstance(img_size, list):
+                if self.is_3D and self.use_3D:
+                    if len(img_size) != 3:
+                        raise ValueError(
+                            "img_size must be declared as a 3-elements list or a single integer value when use_3D is True.")  # noqa
+                else:
+                    if len(img_size) != 2:
+                        raise ValueError(
+                            "img_size must be declared as a 2-elements list or a single integer value when use_3D is False.")  # noqa
+            else:
+                raise ValueError(
+                    "img_size must be declared as a list or a single integer value.")
+
+            # Fix even sizes
+            img_size = [x + 1 if x % 2 != 0 else x for x in img_size]
+
+            self.img_size = img_size
+
             self.imgA_type = self.check_type(self.imgA_paths[0])
             self.imgB_type = self.check_type(self.imgB_paths[0])
             self.is_B_categorical = is_B_categorical
@@ -166,8 +210,24 @@ class DataLoader:
             else:
                 self.norm_boundsB = norm_boundsB
 
-            dataset_property = {"is3D": True,
-                                "input_size": self.input_size,
+            if self.is_3D and self.use_3D and patch_size is not None:
+                if patch_size > img_size[0] or \
+                   patch_size > img_size[1] or \
+                   patch_size > img_size[2]:
+                    raise ValueError(
+                        "patch_size must be lower than img_size.")
+                if patch_overlap < 0.0 or patch_overlap > 1.0:
+                    raise ValueError(
+                        "patch_overlap must be declared as a value between 0 and 1 representing the desidered overlap percentage.")  # noqa
+            else:
+                patch_size = None
+                patch_overlap = 0.0
+
+            self.patch_size = patch_size
+            self.patch_overlap = patch_overlap
+
+            dataset_property = {"is_3D": self.is_3D,
+                                "img_size": self.img_size,
                                 "imgA_label": self.imgA_label,
                                 "imgB_label": self.imgB_label,
                                 "imgA_type": self.imgA_type,
@@ -178,7 +238,9 @@ class DataLoader:
                                 "num_classes": self.num_classes,
                                 "norm_boundsA": self.norm_boundsA,
                                 "norm_boundsB": self.norm_boundsB,
-                                "use_3D": self.use_3D
+                                "use_3D": self.use_3D,
+                                "patch_size": self.patch_size,
+                                "patch_overlap": self.patch_overlap
                                 }
 
             output_dir_content = os.listdir(self.output_dir)
@@ -203,9 +265,9 @@ class DataLoader:
                     json.dump(dataset_property, property_file, indent=2)
 
         elif mode == "get":
-            if not os.path.exists(data_dir):
-                raise FileNotFoundError(f"{data_dir} does not exist")
-            self.output_dir = data_dir
+            if not os.path.exists(data_path) or (not os.path.isdir(data_path)):
+                raise FileNotFoundError(f"{data_path} does not exist")
+            self.output_dir = data_path
 
             # dummy variables for images path
             self.imgA_paths = []
@@ -213,8 +275,8 @@ class DataLoader:
             with open(os.path.join(self.output_dir,
                                    "ds_property.json"), 'r') as property_file:
                 dataset_property = json.load(property_file)
-                self.is_3D = dataset_property["is3D"]
-                self.input_size = dataset_property["input_size"]
+                self.is_3D = dataset_property["is_3D"]
+                self.img_size = dataset_property["img_size"]
                 self.imgA_label = dataset_property["imgA_label"]
                 self.imgB_label = dataset_property["imgB_label"]
                 self.imgA_type = dataset_property["imgA_type"]
@@ -226,6 +288,8 @@ class DataLoader:
                 self.norm_boundsA = dataset_property["norm_boundsA"]
                 self.norm_boundsB = dataset_property["norm_boundsB"]
                 self.use_3D = dataset_property["use_3D"]
+                self.patch_size = dataset_property["patch_size"]
+                self.patch_overlap = dataset_property["patch_overlap"]
 
     def get_dataset(self,
                     batch_size=32,
@@ -311,13 +375,14 @@ class DataLoader:
             ds = ds.map(lambda img: tf.image.rgb_to_grayscale(img),
                         num_parallel_calls=AUTOTUNE)
 
-        # ds = ds.map(lambda img: self.check_dims(img,
-        #                                        self.input_size),
-        #            num_parallel_calls=AUTOTUNE)
-        ds = ds.map(lambda img: self.fix_image_dims(img,
-                                                    self.input_size),
+        ds = ds.map(lambda img: self.fix_image_dims(img),
                     num_parallel_calls=AUTOTUNE)
 
+        if self.patch_size is not None:
+            ds = ds.map(lambda img: self.patch_volume(img),
+                        num_parallel_calls=AUTOTUNE)
+            ds = ds.unbatch()
+
         if is_categorical:
             ds = ds.map(lambda img: tf.one_hot(tf.squeeze(tf.cast(img,
                                                                   img_type)),
@@ -348,9 +413,8 @@ class DataLoader:
                 when images are named with a progressive number inside a folder
                 (e.g.: 001.xxx, 002.xxx, ..., 999.xxx)
         """
-        # print("Fetching images paths...")
-        subset_dir_imgA = os.path.join(self.data_dir, self.imgA_label)
-        subset_dir_imgB = os.path.join(self.data_dir, self.imgB_label)
+        subset_dir_imgA = os.path.join(self.data_path, self.imgA_label)
+        subset_dir_imgB = os.path.join(self.data_path, self.imgB_label)
 
         filenames_imgA = os.listdir(subset_dir_imgA)
         filenames_imgB = os.listdir(subset_dir_imgB)
@@ -360,8 +424,6 @@ class DataLoader:
         paths_imgB = [os.path.join(subset_dir_imgB, img)
                       for img in filenames_imgB]
 
-        # print("Images paths collected.")
-
         # Sort paths alphabetically
         paths_imgA.sort()
         paths_imgB.sort()
@@ -372,6 +434,27 @@ class DataLoader:
 
         return paths_imgA, paths_imgB
 
+    def read_imgs_paths(self):
+        """Read paths of every single image divided by classes from a json file.
+
+        Returns:
+            list, list: two list containing the paths of every images for both
+                classes. The list is sorted alphabetically, this can be usefull
+                when images are named with a progressive number inside a folder
+                (e.g.: 001.xxx, 002.xxx, ..., 999.xxx)
+        """
+        with open(self.data_path, 'r') as f:
+            dataset_paths = json.load(f)
+
+        paths_imgA = dataset_paths[self.imgA_label]
+        paths_imgB = dataset_paths[self.imgB_label]
+
+        if len(paths_imgA) != len(paths_imgB):
+            raise ValueError(
+                f"Dimension mismatch: {len(paths_imgA)} != {len(paths_imgB)}")
+
+        return paths_imgA, paths_imgB
+
     def open_img(self, path):
         """Open an image file and convert it to a tensor.
 
@@ -388,6 +471,7 @@ class DataLoader:
 
         if (self.use_3D):
             image = np.transpose(image, axes=(2, 1, 0))
+            # TODO check the correct orientation for axis
 
         tensor = tf.convert_to_tensor(image)
 
@@ -431,27 +515,58 @@ class DataLoader:
 
         return img_type
 
-    def fix_image_dims(self, img, size):
+    def fix_image_dims(self, img):
         """Fix tensor dimensions so that they are of the
         proper size to carry out Tensorflow operations.
 
         This function performs three steps:
 
-        #. `Squeeze <https://www.tensorflow.org/api_docs/python/tf/squeeze>`_ to remove axis with dimension of 1
-        #. `Expand <https://www.tensorflow.org/api_docs/python/tf/expand_dims>`_ the dimensions of the tensor by adding one axis
-        #. `Resize and pad <https://www.tensorflow.org/api_docs/python/tf/image/resize_with_pad>`_ the tensor to a target width and height
+        # . `Squeeze <https://www.tensorflow.org/api_docs/python/tf/squeeze>`_ to remove axis with dimension of 1
+        # . `Expand <https://www.tensorflow.org/api_docs/python/tf/expand_dims>`_ the dimensions of the tensor by adding one axis
+        # . `Resize and pad <https://www.tensorflow.org/api_docs/python/tf/image/resize_with_pad>`_ the tensor to a target width and height
 
         If `use_3D` was enabled, volume is not resized and padded.
 
         Args:
             img: image or volume to be processed
-            size: desired size of image or volume in the two/three axis.
-
 
         """
-        img = tf.expand_dims(tf.squeeze(img), axis=-1)
-        if (not self.use_3D):
-            img = tf.image.resize_with_pad(img, size, size)
+        # Pad image
+        current_size = tf.shape(img)
+        diff = (self.img_size - current_size) // 2
+        pad_amount = tf.where(diff > 0, diff, 0)
+        pad_amount = tf.expand_dims(pad_amount, axis=-1)
+        paddings = tf.repeat(pad_amount, 2, axis=1)
+        img = tf.pad(img, paddings=paddings)
+
+        # Crop image
+        current_size = tf.shape(img)
+        diff = (self.img_size - current_size) // 2
+        crop_begins = tf.where(diff < 0, -diff, 0)
+        # crop_ends = tf.repeat(self.img_size, tf.shape(crop_begins))
+        crop_ends = self.img_size
+        img = tf.slice(img, crop_begins, crop_ends)
+
+        img = tf.expand_dims(img, axis=-1)
+        return img
+
+    def patch_volume(self, img):
+        overlap_size = int(self.patch_size * self.patch_overlap)
+        ksizes = [1,
+                  self.patch_size,
+                  self.patch_size,
+                  self.patch_size,
+                  1]
+        strides = [1,
+                   self.patch_size - overlap_size,
+                   self.patch_size - overlap_size,
+                   self.patch_size - overlap_size,
+                   1]
+        img = tf.expand_dims(img, axis=0)
+        img = tf.extract_volume_patches(img, ksizes, strides, 'VALID')
+        img = tf.reshape(img,
+                         [-1, self.patch_size, self.patch_size, self.patch_size])
+        img = tf.expand_dims(img, axis=-1)
         return img
 
     # -------------------------------------------------------------------------
@@ -499,10 +614,10 @@ class DataLoader:
         return imgA, imgB
 
 
-def generate_dataset(data_dir,
+def generate_dataset(data_path,
                      imgA_label,
                      imgB_label,
-                     input_size,
+                     img_size,
                      output_dir=None,
                      extract_only=None,
                      norm_boundsA=None,
@@ -510,20 +625,24 @@ def generate_dataset(data_dir,
                      is_B_categorical=False,
                      num_classes=None,
                      use_3D=False,
+                     patch_size=None,
+                     patch_overlap=0,
                      ):
 
     data_loader = DataLoader(mode="gen",
-                             input_size=input_size,
+                             data_path=data_path,
+                             img_size=img_size,
                              imgA_label=imgA_label,
                              imgB_label=imgB_label,
-                             data_dir=data_dir,
                              output_dir=output_dir,
                              is_B_categorical=is_B_categorical,
                              num_classes=num_classes,
                              norm_boundsA=norm_boundsA,
                              norm_boundsB=norm_boundsB,
                              extract_only=extract_only,
-                             use_3D=use_3D
+                             use_3D=use_3D,
+                             patch_size=patch_size,
+                             patch_overlap=patch_overlap,
                              )
 
     data_loader.get_dataset()
@@ -547,7 +666,7 @@ def get_dataset(data_dir,
         raise ValueError("Sum of percentages has to be 1")
 
     data_loader = DataLoader(mode="get",
-                             data_dir=data_dir,
+                             data_path=data_dir,
                              )
 
     complete_ds = data_loader.get_dataset(batch_size=batch_size,
@@ -560,7 +679,7 @@ def get_dataset(data_dir,
 
     # Compute length of dataset
     num_imgs = 0
-    for img in complete_ds:
+    for _ in complete_ds:
         num_imgs += 1
 
     train_ends = int(num_imgs * percentages[0])
